/* $Id: vircam_mods.h,v 1.38 2013-10-15 16:48:00 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2015 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-15 16:48:00 $
 * $Revision: 1.38 $
 * $Name: not supported by cvs2svn $
 */

#ifndef VIRCAM_MODS_H
#define VIRCAM_MODS_H

#include <cpl.h>
#include "casu_fits.h"
#include "casu_tfits.h"
#include "casu_mask.h"

/* Function prototypes */

extern int vircam_destripe(casu_fits *in, casu_mask *inbpm, int *status);

extern int vircam_genbpm(casu_fits **flatlist, int nflatlist, float lthr,
			 float hthr, cpl_array **bpm_array, int *nbad,
			 float *badfrac, int *status);

extern int vircam_genlincur(double **fdata, int nimages, double *exps,
			    double mindit, casu_tfits *chantab, 
			    int norder, cpl_table **lchantab, 
			    double **lindata, int *status);

extern int vircam_grout(casu_tfits *intab, cpl_frameset *input_cats, 
			cpl_frameset *input_confs, casu_tfits **outtab,
			int *status);


extern int vircam_interleave(casu_fits **infiles, int ninputs, 
			     casu_fits **inconfs, int nconfs, int nsteps,
			     cpl_propertylist **p, cpl_image **outimage, 
			     cpl_image **outconf, int *status);

extern int vircam_lincor(casu_fits *infile, casu_tfits *lchantab,
			 int kconst, int ndit, int *status);

extern int vircam_mosaic(cpl_frameset *infiles, cpl_frameset *inconf,
			 int interp, int skyflag, float skywish_in, 
			 const char *expkey, int conflim, casu_fits **out,
			 casu_fits **outc, int *status);

extern int vircam_platesol(cpl_propertylist *plist, cpl_propertylist *tlist,
			   cpl_table *matchedstds, int nconst, int shiftan, 
			   int *status);


#endif

/*

$Log: not supported by cvs2svn $
Revision 1.37  2011/05/06 11:26:03  jim
removed extraneous declaration

Revision 1.36  2009/01/19 14:35:23  jim
Added vircam_nditcor

Revision 1.35  2008/11/21 10:11:12  jim
Added vircam_imstack and vircam_platexy

Revision 1.34  2008/10/24 10:56:08  jim
Modified vircam_tpoffset parameters

Revision 1.33  2008/10/21 08:40:51  jim
Added vircam_tpoffset

Revision 1.32  2008/10/13 08:16:10  jim
Added vircam_opm and vircam_inpaint

Revision 1.31  2007/11/26 09:57:39  jim
vircam_lincor now take account of ndit

Revision 1.30  2007/11/22 12:34:46  jim
Modifed calling parameters for vircam_genlincur

Revision 1.29  2007/11/14 10:44:28  jim
Modified to add vircam_genbpm and to rewrite the entry for vircam_genlincur

Revision 1.28  2007/10/25 17:34:00  jim
Modified to remove lint warnings

Revision 1.27  2007/05/08 10:42:00  jim
Added vircam_gaincor

Revision 1.26  2007/05/02 09:13:29  jim
Modified to change api of vircam_imcore (outtab is now vir_tfits) and
vircam_platesol (takes table propertylist)

Revision 1.25  2006/12/01 14:10:35  jim
Added vircam_defringe

Revision 1.24  2006/11/28 22:10:22  jim
Added illcor_rms to calling sequence of vircam_illum

Revision 1.23  2006/11/28 20:53:45  jim
Added vircam_illum

Revision 1.22  2006/11/27 12:04:02  jim
Modified prototype for vircam_getstds

Revision 1.21  2006/10/02 13:44:04  jim
Added vircam_destripe

Revision 1.20  2006/08/01 11:27:53  jim
Modifications to imcore background estimation and to add ability to
specify the smoothing kernel width

Revision 1.19  2006/05/30 14:27:54  jim
Added illcor parameter to call for photcal

Revision 1.18  2006/05/17 12:06:17  jim
Added vircam_photcal

Revision 1.17  2006/05/15 13:15:28  jim
added vircam_imdither

Revision 1.16  2006/04/20 11:28:29  jim
Added vircam_platesol and a few minor mods

Revision 1.15  2006/03/22 13:19:29  jim
modified vircam_mkconf

Revision 1.14  2006/03/22 12:02:55  jim
modified vircam_mkconf prototype

Revision 1.13  2006/03/15 10:43:41  jim
Fixed a few things

Revision 1.12  2006/03/08 14:32:22  jim
Lots of little modifications

Revision 1.11  2006/03/06 13:49:08  jim
Modified so that the DRS keywords are written directly to the extension header
for the input image

Revision 1.10  2006/03/03 14:29:46  jim
Modified definition of casu_fits and channel table

Revision 1.9  2006/03/01 10:31:29  jim
Now uses new vir_fits objects

Revision 1.8  2006/02/22 10:09:33  jim
Added status variables to vircam_imcombine and vircam_imcore

Revision 1.7  2006/02/18 11:48:55  jim
*** empty log message ***

Revision 1.6  2006/01/23 22:58:41  jim
*** empty log message ***

Revision 1.5  2006/01/23 10:30:31  jim
Modifications mainly to docs

Revision 1.4  2006/01/03 10:30:22  jim
Modified vircam_darkcor and vircam_flatcor

Revision 1.3  2005/10/14 13:18:25  jim
Modified parameter list for vircam_imcombine

Revision 1.2  2005/09/20 15:07:46  jim
Fixed a few bugs and added a few things

Revision 1.1.1.1  2005/08/05 08:29:09  jim
Initial import


*/
